`timescale 10ns / 10ps

`define SOUND_MAX	8'hFF

module alx_audio(
	input 				bSys80MClock,				// System clock 
	input	[7:0]			V8_DigInput,				// 8 bit input
	input 				rst_n,						// reset
	output reg			oPWM_P						// PWM out	
);

reg [7:0]  V8_PWM_P_Cur;       					// State machine and current value of sound PWM 
reg [7:0]  V8_PWM_P_Target;    					// Target for sound PWM                         


always @(posedge bSys80MClock or negedge rst_n)
begin
    if (rst_n == 0)                          // If we are in reset                                           
    begin
        oPWM_P <= 0;                         // Output 0 to PWMp out                                         
        V8_PWM_P_Cur = 0;                    // Current PWM output                                           
        V8_PWM_P_Target = 0;                 // PWM target from SNOR reader                                  
    end else
    begin
        case (V8_PWM_P_Cur)                  // Check on current PWM state                                   
        255:                                 // End of the cycle                                             
        begin
            V8_PWM_P_Target = V8_DigInput;   // Load PWM target from SNOR reading routine                    
            if (V8_DigInput == `SOUND_MAX)   // If 0xFF is being output                                          
            begin 
                oPWM_P <= 1;                 // PWMp set to 1                                                
                V8_PWM_P_Cur = 0;            // Reset state machine                                          
            end else
            begin
                oPWM_P <= 0;                 // PWMp set to 0                                                
                V8_PWM_P_Cur = 0;            // Reset state machine                                          
            end
        end
        default:                             // Mid of the PWM cycle                                         
        begin
                                             // If target has been reached and not max output being set      
            if (V8_PWM_P_Cur >= V8_PWM_P_Target && V8_PWM_P_Target != `SOUND_MAX)
            begin 
                oPWM_P <= 0;                 // output 0 as PWMp out                                         
															// Increment state machine                                       
					 V8_PWM_P_Cur = V8_PWM_P_Cur + 1;       
            end else
            begin
                oPWM_P <= 1;                 // output 1 as PWMp output                                      
                										// Increment state machine                                       
					 V8_PWM_P_Cur = V8_PWM_P_Cur + 1;       
            end
        end
        endcase
    end
end

endmodule