`timescale 10ns / 10ps

`define VSYNCSTART			0
`define VSYNCEND				2
`define HSYNCSTART			0
`define HSYNCEND				96
`define HTOTAL					864
`define HBACKPORCH			64
`define VTOTAL					600
`define HACTSTART				144
`define HACTEND				784
`define VACTSTART				35
`define VACTEND				515
`define VBACKPORCH			48
`define VACTPAPER				467
`define VIDEO_PIX_INT		0
`define FLASHSPEED			19
`define ZX_RAM_START			16384
`define ZX_MAIN_MEM			32768
`define STARTUP_DELAY 		3
`define COMMAND_SIZE 		32
`define READ_COMMAND 		8'h03
`define DATA_SIZE 			8
`define RAM_OFFSET			32768
`define ROM2ND_OFFSET		16384

module aZXSpectrum
(
	input 				key1,							// Debug key
	input 				key2,							// Debug key
	input 				key3,							// Debug key
	input           	clk,           			// system clock 50Mhz on board
	input           	rst_n,         			// reset, low active
	input	 	 			keyb_in,						// keyb_in, keyboard in
	input		 			tapein, 						// ZX Spectrum tape in
	input 				turbo_mode,					// Speed selector input for tape loader
	input     [4:0]	KempstonJoystick,			// Kempston Joystick
	input             iSNOR_DO, 					// nDO, pin2 of 25Q16 *******
	output reg[7:0]	tstDATA_LED,				// 7 seg LED data
	output reg[5:0]	tstDATA_LED_SEL,			// 7 seg LED select
	output reg			oSNOR_nCS,					// nCS, pin1 of 25Q16 *******
	output reg        oSNOR_DI, 					// nDI, pin5 of 25Q16 *******
	output reg        oSNOR_CLK, 					// nCK, pin6 of 25Q16 *******
	output  				vga_hsync,					// Out h sync
	output  				vga_vsync,    				// Out v sync
	output reg 			ex_nWR_EN,					// Write enable
	output reg 			ext_n_CE,					// External RAM chip enable
	output reg 			ext_n_OE,					// External RAM output enable
	output reg 			sound,						// ZX Spectrum sound
	output reg[3:0] 	led,           			// LED, use for control the LED signal on board
	output reg[4:0]	out_R,						// Output RED
	output reg[5:0]	out_G,						// Output GREEN
	output reg[4:0]	out_B,						// Output BLUE
	output reg[18:0]  extADDR,						// Address bus to an external RAM
	inout     [7:0]   w_extDATA,					// Data bus for external RAM
	output 				audio_out					// 8912 output
);
															// ******************Video variables*******************
reg		  outVideoPCLK;							// VGA pclk counter
reg		  [9:0]outPixCntr;						// Pixels in line counter
reg		  [9:0]outLineCntr;						// Lines in frame counter
reg		  outVsync;									// Output vsync
reg 		  outHsync;									// Output hsync
reg [23:0] pixeladdr;								// Pixel address before transferring into ZX format

reg [12:0] aVRAM_addr_a;							// Bank 5, Video memory address - for video adapter
reg [12:0] aVRAM_addr_b;							// Bank 5, Video memory address - for Z80
reg [7:0]  aVRAM_data_a;							// Bank 5, Video data IN - for video adapter, not used
reg [7:0]  aVRAM_data_b;							// Bank 5, Video data IN - for Z80
wire [7:0] aVRAM_q_a;								// Bank 5, Video data OUT - for video adapter		
wire [7:0] aVRAM_q_b;								// Bank 5, Video data OUT - for Z80
reg        aVRAM_wren_a;							// Bank 5, Video data, Write enable - for video adapter
reg        aVRAM_wren_b;							// Bank 5, Video data, Write enable - for Z80

reg [12:0] bVRAM_addr_a;							// Bank 7, Video memory address - for video adapter
reg [12:0] bVRAM_addr_b;							// Bank 7, Video memory address - for Z80
reg [7:0]  bVRAM_data_a;							// Bank 7, Video data IN - for video adapter, not used
reg [7:0]  bVRAM_data_b;							// Bank 7, Video data IN - for Z80
wire [7:0] bVRAM_q_a;								// Bank 7, Video data OUT - for video adapter		
wire [7:0] bVRAM_q_b;								// Bank 7, Video data OUT - for Z80
reg        bVRAM_wren_a;							// Bank 7, Video data, Write enable - for video adapter
reg        bVRAM_wren_b;							// Bank 7, Video data, Write enable - for Z80

reg [6:0]  ColorLineAddr;							// Color attribute counter (1 line only), range 0..31, used in calculating VRAM memory address
reg [255:0]ColorDataMemCpy;						// Fetched data from color memory for 1 line only
reg [7:0]  CurColorAttribute;						// Current color attribute
reg [6:0]  CurColorLineAddr;						// Address for an attrubute cache
reg 		  ColorSync;								// Signal to make proper timing to fetch data from FPGA RAM
reg 		  rZXFlashAttr;							// Flash-attribute
reg [4:0]  rZXFlashCntr;							// Counter for Flash attribute

wire[15:0] wZXGlobalAddress;						// Global Z80 memory address
reg [7:0]  wZXGlobalDataO;							// Global memory data out

reg [7:0]  rExternal_mem_data_bus;				// External RAM data bus
															// External RAM wr/rd
assign w_extDATA = ex_nWR_EN ? 8'bZZZZZZZZ : rExternal_mem_data_bus;

reg[15:0]  rLED_Mcntr;								// LED timer
reg[23:0]  rLED_Data;								// Hex data to be displayed
reg [3:0]  rLED_Bcd;									// Single digit hex to be displayed	
reg[2:0]	  rLED_Sel;									// Binary LED selector
reg		  rLED_Special;							// Special set of characters meaning to mimic MK61 

reg [2:0] rBorderColor;								// Border color

reg [7:0] r8Uart115RxState;                  // RX state machine                             
reg [7:0] r8Uart115RxNextState;              // RX next state                                
reg [7:0] r8Uart115RxDelay;                  // RX delay                                     
reg [3:0] r4Bits115RxCount;                  // RX bits counter                              
reg [7:0] r8Uart115RxByte;                   // RX byte                                          
reg		 bRx115ShiftReq;                    // Bit shift request                            
reg       bRXD115_ready;                     // RX data ready                                
reg [7:0] rUdata115Out;                      // Received byte                                
reg clk25;												// 25M clock
reg m28clk;
wire r3p5clk;											// 3p5MHz clock for CPU
reg bkeyb_in;											// Register to hold input value for keyboard input to prevent internal overload
reg [7:0]rx_byte_cntr;								// Byte counter of external stream

reg [3:0] clk_div_cntr;				   			// Divider counter

reg [39:0]zx_Keyboard_state;						// State of ZX keyboard

/* Z80 pins */
wire 		a_nM1;										// It indicates the Machine Cycle number 1. So the CPU activates this output, when 
															// the first machine cycle of an instruction is in progress
															
wire 		a_nMREQ;										// The memory request pin used to return low (0) signal, when communication with memory blocks is desired.
wire 		a_nIORQ;										// The Input Output request pin is used to return low (0) signal when communication with IO ports is desired.
wire 		a_nRD;										// Read signal
wire 		a_nWR;										// Write signal
wire 		a_nRFSH;										// The RFSH and MREQ both can be used to indicate the lower seven bits of the address bus can be used as 
															// refresh address for Dynamic memories.
															
wire 		a_nHALT;										// The Halt state indicates the CPU has executed a HALT instruction and waiting for interrupt.
wire 		a_nBUSACK;									// It is Bus Acknowledge pin. It indicates the requesting device that the address bus, data bus, 
															// control bus enter into a high impedance state.

reg 		a_nWAIT;										// This signal is used for the communication between CPU and memory units or IO devices. Until the signal 
															// is inactive the CPU continuously enters the Wait state.
															
reg 		a_nINT;										// Interrupt Request pin is used to generate interrupts by IO devices. After executing each instruction, CPU 
															// checks for the Interrupt requests
reg 		a_nNMI;										// The Non-Maskable Interrupt pin is negative edge triggered. Its priority is higher than the INT. It does 
															// not depend on the status of the interrupt enable flip-flops. It automatically forces the CPU to 
															// restart at location 0066H.

reg 		a_nBUSRQ;									// Bus request Pin forces the address bus, data bus and control bus to enter a high impedance state. 
															// It is always recognized at the end of each machine cycle.

wire [7:0] a_D;										// Data bus for Z80 CPU

															// Assign signal for Z80 which are not used in this implementation of ZX Spectrum
assign vga_hsync = outHsync;						// Assign output hsync pin to signal name
assign vga_vsync = outVsync;        			// Assign output vsync pin to signal name

reg [31:0] sp50Hzcntr;								// Counter of clocks to generate an interrupt

reg [7:0] r8Key1;										// First pressed key
reg [7:0] r8Key2;										// Second pressed key
reg [7:0] r8Key3;										// Third pressed key


reg bDrawingBorder;									// Drawing border flag
reg PreviousVsync;

wire rCpuClkMain;
reg rClkSwitch;									   // Clock select signal
reg rClkSwitchPrev;									// Previous state of clock select
reg [23:0] rSwDelay;									// Delay to minimize glitches during clock switch
// ************** SPI reader 
reg [3:0]  V4_MainState;       					// Main state machine                           
reg [7:0]  V8_Misc;            					// Miscelanious storage                         
reg [7:0]  V8_DIr;             					// Data read from SPI bus                       
reg [23:0] V24_StartAddr;      					// Start address of sound location in SNOR      
reg [23:0] V24_EndAddr;        					// End address of sound location in SNOR        
reg [31:0] V32_Misc;          					// Secondary variable

reg [31:0] v32CheckSum;

reg bZ80Reset;											// Reset for Z80			
reg bSPI_done;											// Bit indication that SPI readback is done
reg bSPIind;											// Bit indicating that trasfer for given byte is over via SPI
reg bPrevSPI; 											// Bit indicating previou state of SPI transfer

reg [18:0] aROMCopyAddr;							// An address for copying ROM from SNOR to RAM
reg [2:0]  b2RomCopySM;								// State machine for ROM copying

reg [7:0] r8_128KSwitch;							// Port 7FFD to manage 128K

reg [7:0] AY8910_d_in;								// AY 8910 data input
wire[7:0] AY8910_d_out;								// AY 8910 data output
wire		 AY8910_d_oe_l; 							// Output signal (not used)
wire 		 AY8910_BDIR;								// BDIR input
wire 		 AY8910_BC1;								// BC 1 input
wire [7:0]wSound8b;									// 8 bit sound
wire		 c80M;										// 80M clock for audio pwm
reg 		 aud_half_clock;							// 1.75MHz clock for audio	

/* 

The following logic has been implemented for AY8910:

BDIR = not (a1 or nWR) provided that nIRQ = 0 
BDIR = 0 if nIRQ = 1

BC1 = not (a1 or not a14) provided that nIRQ = 0
BC1 = 0 if nIRQ = 1

*/

assign AY8910_BDIR  = a_nIORQ ? 0 : !(wZXGlobalAddress[1] | a_nWR);
assign AY8910_BC1 = a_nIORQ ? 0 : !(wZXGlobalAddress[1] | !wZXGlobalAddress[14]);

															// CPU clock switch based on the signal from tape loader
assign rCpuClkMain = rClkSwitch? r3p5clk:m28clk;
			
pll_rep35M aPLL35 (									// PLL for clocks	
	.inclk0(clk),										// Input 50MHz
	.c0(r3p5clk),										// Ouput normal CPU clock, 3.5MHz
	.c1(m56clk)											// Output 56MHz
);

vram_data aVRAM(										// Bank 5, Video RAM, no color
.address_a(aVRAM_addr_a), 							// Bank 5, Video memory address - for video adapter
.address_b(aVRAM_addr_b),							// Bank 5, Video memory address - for Z80
.clock(m56clk),										// Bank 5, 56MHz clock
.data_a(aVRAM_data_a),								//	Bank 5, Video data IN - for video adapter, not used	
.data_b(aVRAM_data_b),								// Bank 5, Video data IN - for Z80
.wren_a(aVRAM_wren_a),								// Bank 5, Video data, Write enable - for video adapter
.wren_b(aVRAM_wren_b),								// Bank 5, Video data, Write enable - for Z80
.q_a(aVRAM_q_a),										// Bank 5, Video data OUT - for video adapter	
.q_b(aVRAM_q_b)										// Bank 5, Video data OUT - for Z80
);

vram_data bVRAM(										// Bank 7, Video RAM, no color
.address_a(bVRAM_addr_a), 							// Bank 7, Video memory address - for video adapter
.address_b(bVRAM_addr_b),							// Bank 7, Video memory address - for Z80
.clock(m56clk),										// Bank 7, 56MHz clock
.data_a(bVRAM_data_a),								//	Bank 7, Video data IN - for video adapter, not used	
.data_b(bVRAM_data_b),								// Bank 7, Video data IN - for Z80
.wren_a(bVRAM_wren_a),								// Bank 7, Video data, Write enable - for video adapter
.wren_b(bVRAM_wren_b),								// Bank 7, Video data, Write enable - for Z80
.q_a(bVRAM_q_a),										// Bank 7, Video data OUT - for video adapter	
.q_b(bVRAM_q_b)										// Bank 7, Video data OUT - for Z80
);

T80a aZ80(
		.RESET_n(bZ80Reset),					 		// Reset 
		.CLK_n(rCpuClkMain),							// Clock input
		.WAIT_n(a_nWAIT),								// WAIT input
		.INT_n(a_nINT),								// INT input
		.NMI_n(a_nNMI),								// NMI input
		.BUSRQ_n(a_nBUSRQ),							// BUSRQ INPUT
		.M1_n(a_nM1),									// M1 output
		.MREQ_n(a_nMREQ),								// MREQ output
		.IORQ_n(a_nIORQ),								// IORQ output
		.RD_n(a_nRD),									// RD output
		.WR_n(a_nWR),									// WR output
		.RFSH_n(a_nRFSH),								// RFSH output
		.HALT_n(a_nHALT),								// HALT output
		.BUSAK_n(a_nBUSACK),							// BUSACK output
		.A(wZXGlobalAddress),						// Address outout
		.D(a_D)											// Data inout
);

YM2149 AY8910(
  .I_DA(AY8910_d_in),                			// Audio data in
  .O_DA(AY8910_d_out),       		         	// Audio data out
  .I_A9_L(0),              						// A9_L
  .I_A8(1),                						// A8
  .I_BDIR(AY8910_BDIR),              			// BDIR
  .I_BC2(1),               						// BC2
  .I_BC1(AY8910_BC1),               			// BC1
  .I_SEL_L(1),             						// SEL_L

  .O_AUDIO(wSound8b),             				// Ouput
  .ENA(aud_half_clock),                 		// CE signal
  .RESET_L(rst_n),             					// reset
  .CLK(m28clk)                 					// 28MHz
  );

alx_audio alx_out(
	.bSys80MClock(c80M),								// System clock 
	.V8_DigInput(wSound8b),							// 8 bit input
	.rst_n(rst_n),										// reset
	.oPWM_P(audio_out)								// PWM out	
);

clocks AudioClkGen(									// Audio clocks generator
	.CLK(m28clk),										//:	28M in
	.nRESET(rst_n),									//:	1.75 out
	.CLKEN_PSG(aud_half_clock)						//:	
	);

pll80 alx_snd_pll(
	.inclk0(clk),										// Input 50MHz
	.c0(c80M)											// Ouput 80MHz
);  
															
															// Assignment of data bus based on a_nWR signal		
assign a_D = a_nRD ? 8'bZZZZZZZZ:wZXGlobalDataO;
															// Polling clock switch to debounce turbo mode signal
always @ (posedge rCpuClkMain or negedge rst_n)	
begin
	if (rst_n == 0)									// Reset is active?
	begin
		rClkSwitch <= 0;								// If yes, then reset all of the signals
		rSwDelay <= 0;									// Reset debounce
		rClkSwitchPrev <= 0;							// Reset previous state of selector
	end else
	begin
		if (rSwDelay == 16'hFFFF)					// If delay is expired (debounce)
		begin
															// If no RAM/ROM access and switch position changed
			if (a_nRD == 1 && a_nWR == 1 && rClkSwitchPrev != turbo_mode)
			begin
				rClkSwitchPrev <= turbo_mode;    // Set prvious state of turbo selector
				rSwDelay <= 0;							// Set debounce timer
				if (turbo_mode == 0)					// If turbo switch is off
				begin
					rClkSwitch <= 0;					// Then setting switch variable to 0
				end else
				begin
					rClkSwitch <= 1; 					// Otherwise setting switch variable to 1
				end
			end
		end else
		begin
			rSwDelay <= rSwDelay + 1;				// Increment debounce delay
		end
	end
end

always @ (posedge m56clk or negedge rst_n)	// Generate 28M clk for CPU in turbo mode
begin
	if (rst_n == 1'b0)   	                  // If reset, then initialize everything         
	begin
		m28clk <= 0;									// Initialize clock as 0 at reset
	end else
	begin
		m28clk = ~m28clk;								// Flip the bit every 56M clock pulse, so it becomes 28M
	end
end

always @(negedge r3p5clk or negedge rst_n) 	// SPI reader
begin
	if (rst_n == 0)									// If Reset = 0 then it means reset is active                   
	begin
		oSNOR_nCS <= 1;								// At reset nCS = 1                                             
		oSNOR_CLK <= 1;								// At reset CLK = 1                                             
		oSNOR_DI <= 1;									// At reset DI = 1                                              
		V4_MainState = 0;								// Initialize main state machine counter                        
		V32_Misc = 0;									// Initialize secondary counter                                 
		V8_Misc = 0;									// Miscelanious storage                                         
		V8_DIr = 0;										// Read data value eq 0 at reset                                
		V24_StartAddr = 0;							// Play first sound at the beginning                            
		V24_EndAddr = 0;								// End of first sound                                           
		bSPI_done = 0;									// Bit pointing out that SPI transfer is done
		bSPIind = 0;									// Bit indicating that trasfer for given byte is over via SPI
	end else
	begin
		case (V4_MainState)							// Main state machine				
			0:												// Startup delay to ensure that everything is powered up        
			begin
				if (V32_Misc < `STARTUP_DELAY )	// If MISC counter less than startup delay (0xFFFF)             
				begin
					V32_Misc = V32_Misc + 1;		// Increment MISC counter                                       
				end else
				begin
					V4_MainState = 8;					// Otherwise go to init state                                   
					V32_Misc = 0;						// Set misc variable as NULL                                    
				end
			end
			1:												// prepare to send over SPI                                     
			begin
				oSNOR_nCS <= 0;						// Set nCS = 0                                                  
				oSNOR_CLK <= 0;						// Set CLK = 0                                                  
				V32_Misc[31:24] = `READ_COMMAND;	// 03/READ_COMMAND is the cmd to read followed by 24b address   
				V32_Misc[23:0] = V24_StartAddr;	// Load start address                                           
				V8_Misc = `COMMAND_SIZE;			// Load command size                                            
				V4_MainState = V4_MainState + 1;	// Go to next state                                             
			end
			2:												// Setting DATA bit on SPI                                      
			begin
				oSNOR_DI <= V32_Misc[31];			// Set SNOR_DI to be equal first bit to sent                    
				V4_MainState = V4_MainState + 1;	// Next state                                                   
			end
			3:												// Internal shift left and setting SPI CLK                      
			begin
				V32_Misc[31:1] = V32_Misc[30:0];	// Shift left by 1 bit                                          
				oSNOR_CLK <= 1;						// SNOR_CLK = 1                                                 
				V4_MainState = V4_MainState + 1;	// Next state                                                   
				V8_Misc = V8_Misc - 1;				// Decrease amount of bits to be sent                           
			end 
			4:												// Clearing the SPI CLK and checking if 8 bits are sent out     
			begin
				oSNOR_CLK <= 0;						// Set SNOR_CLK = 0                                             
				if (V8_Misc)							// If this variable is not 0 then there bits left to be sent    
				begin
					V4_MainState = 2;					// Go back to state 2 to send more                              
				end else
				begin
					V4_MainState=V4_MainState + 1;// All bits are sent, go to next state                          
					V8_Misc = `DATA_SIZE;			// Size of the data packet                                      
				end
			end
			5:												// Reading the bit from SPI. Data is being set on the falling   
															// edge of the SPI clock                                        
			begin
				V8_DIr[0] = iSNOR_DO;				// Get data bit from the bus                                    
				V4_MainState = V4_MainState + 1;	// Next state                                                   
				oSNOR_CLK <= 1;						// Clock = 1                                                    
			end
			6:												// Clearing the SPI clk and decresing bit count remaining to be 
															// read                                                         
			begin 
				oSNOR_CLK <= 0;						// Clock = 0                                                    
				V4_MainState = V4_MainState + 1;	// Next state                                                   
				V8_Misc = V8_Misc - 1;				// Bit count decrease                                           
			end
			7:												// Checking if all 8 bits are received. If no, the shift the    
															// data and jump back to read state. If yes, then go ahead      
			begin
				if (V8_Misc)							// If set, then not 8 bits has been received yet                
				begin
					V8_DIr[7:1] = V8_DIr[6:0];		// Shift left                                                   
					V4_MainState = 5;					// Jump to next bit received                                    
				end else
				begin
					oSNOR_nCS <= 1;					// Byte read is over, set CS                                    
															// If last address is not reched yet                            
					if (V24_StartAddr != V24_EndAddr)      
					begin
															// Proceed to the next byte                                     
						V24_StartAddr = V24_StartAddr + 1;
						V4_MainState = 1;				// Main state = beginning                                       
						bSPIind = !bSPIind;
					end else 
					begin
						bSPI_done = 1;
					end
				end
			end
			8:
			begin
				V4_MainState = 1;						// Jump back to the beginning                                   
				if (key1 == 0) 
				begin
					V24_StartAddr = 24'h008000;	// Second file                                                   
					V24_EndAddr = 24'h00C000;		// Second file
				end else
				begin
					V24_StartAddr = 24'h000000;	// First file                                                   
					V24_EndAddr = 24'h008000;		// First file
				end
			end
		endcase
	end
end

//rLED_Data is 24 bit input to LED
//rLED_Special is a bit indicating special characters, if this bit is set: 
//\___	0000:  "-" Special thanks to MK61 and MK52
//			0001:  "8" Special thanks to MK61 and MK52
//			0010:  "C" Special thanks to MK61 and MK52
//			0011:  "E  Special thanks to MK61 and MK52
//			0100:  " " Special thanks to MK61 and MK52
//			0101:  "r" Special thanks to MK61 and MK52
//			0110:  "o" Special thanks to MK61 and MK52
//
// Memory interface signals:
//			wZXGlobalAddress 
//    	wZXGlobalDataO

always @ (posedge clk25 or negedge rst_n)		// Poll in the data for USB keyboard in order to go from signal to register
begin
	if (rst_n == 1'b0)   	                  // If reset, then initialize everything         
	begin
		bkeyb_in <= 0;									// Startup value
	end else
	begin
		bkeyb_in <= keyb_in;							// Polling
	end
end

always @(negedge clk25 or negedge rst_n)		// Uart RX procedure for keyboard @115200
begin
	if (rst_n == 1'b0)   	                  // If reset, then initialize everything         
	begin
		r8Uart115RxState <= 0;                 // RX state machine                             
		r8Uart115RxNextState <= 0;             // RX next state                                
		r8Uart115RxDelay <= 0;                 // RX delay                                     
		r4Bits115RxCount <= 0;                 // RX bits counter                              
		r8Uart115RxByte <= 0;                  // RX byte                                          
		bRx115ShiftReq <= 0;                   // Bit shift request                            
		bRXD115_ready <= 0;                    // RX data ready  
		rUdata115Out <= 0;                     // Received byte      
		rx_byte_cntr <= 0;							// Received byte counter
		zx_Keyboard_state <= 40'hffffffffff;	// 40 bit zx keyboard state, see keys.ods
		r8Key1 <= 0;									// First key
		r8Key2 <= 0;									// Second key
		r8Key3 <= 0;									// Third key
	end else
	begin
		case (r8Uart115RxState)                // Main state machine                           
			0:                                  // Default state, waiting for data              
			begin
				if (bkeyb_in == 0)               // if RX is 0 then it means that data is coming 
				begin
					r8Uart115RxNextState <= 1;    // Set next state var to next state             
					r8Uart115RxState <= 255;      // Jump to delay state next                     
					bRx115ShiftReq <= 0;          // There is no need to shift now                
					r8Uart115RxDelay <= 254;      // Set delay counter to 254(115200 baud)        
					r4Bits115RxCount <= 0;        // Reset bit counter                            
					r8Uart115RxByte <= 0;         // Clear RX byte                                
					bRXD115_ready <= 0;           // Clear ready bit   
				end else
				begin
					bRXD115_ready <= 0;           // Otherwise just clear ready bit   
				end
			end
			1:                                  // Receiving in progress                        
			begin
				if (r4Bits115RxCount < 7)        // If we still have some room to received       
				begin			
															// Increment bits counter                       
					r4Bits115RxCount <= r4Bits115RxCount + 1;
					bRx115ShiftReq <= 1;          // Shift will be required                       
					r8Uart115RxDelay <= 216;      // Set delay to 216                             
					r8Uart115RxState <= 255;      // Set state machine to delay state             
					r8Uart115RxNextState <= 1;    // Set state to jump to 1                       
					r8Uart115RxByte[7]<= bkeyb_in;// Get the bit                                  
				end else
				begin                            // Otherwise this is last bit                   
                                             // rUdataOut - is a received byte               
					rUdata115Out[7] <= bkeyb_in;  // Store it                                         
                                             // And store the rest of the bits               
					rUdata115Out[6:0] <= r8Uart115RxByte[6:0]; 
					r8Uart115RxState <= 2;        // Jump to next state                           
				end
			end
			2:                                  // State to set signal that byte is came        
			begin
				bRx115ShiftReq <= 0;             // Clear shift req                                   
				bRXD115_ready <= 1;              // Set the RX data ready bit   
															// If we are within the packet from keyboard
				if (rUdata115Out != 8'h57 && rUdata115Out != 8'hab && rUdata115Out != 8'h82 && rUdata115Out != 8'ha3)	
				begin
					if (rx_byte_cntr == 3)			// Shifts on the keyboard are presented as 3rd byte (count starts form 0)
					begin
															// Display shifts as byte 0
															// Caps shift
						if (rUdata115Out[3:0] == 4'h02)
						begin
															// 0 - if key is pressed
							zx_Keyboard_state[35] <= 0;
						end else
						begin
															// 1 - if key is released											
							zx_Keyboard_state[35] <= 1;
						end 
															// Symbol shift
						if (rUdata115Out[7:4] == 4'h02)
						begin
															// 0 - if key is pressed										
							zx_Keyboard_state[31] <= 0;
						end else
						begin
															// 1 - if key is released											
							zx_Keyboard_state[31] <= 1;
						end 
					end
					
					if (rx_byte_cntr == 5) 			// First pressed key comes as #5 in the sequence of bytes received
					begin	
															// Copy it into the appropriate memory loaction
						r8Key1 <= rUdata115Out[7:0];
					end
					
					if (rx_byte_cntr == 6) 			// Second pressed key comes as #6 in the sequence of bytes received
					begin
						r8Key2 <= rUdata115Out[7:0];
															// Copy it into the appropriate memory loaction
					end

					if (rx_byte_cntr == 7) 			// Third pressed key comes as #7 in the sequence of bytes received
					begin
															// Copy it into the appropriate memory loaction
						r8Key3 <= rUdata115Out[7:0];
					end
															// Actual keys presented as 5th/6th/7th byte (count starts from 0). So 8th state is when sequence complete
					if (rx_byte_cntr == 8)
					begin									
															// ZXCV
															// Z key, 0 pressed and 1 is released
						if (r8Key1 == 8'h1d || r8Key2 == 8'h1d || r8Key3 == 8'h1d) 
						begin
							zx_Keyboard_state[36] <= 0;
						end else
						begin
							zx_Keyboard_state[36] <= 1;
						end 
															// X key, 0 pressed and 1 is released	
						if (r8Key1 == 8'h1b || r8Key2 == 8'h1b || r8Key3 == 8'h1b) //X
						begin
							zx_Keyboard_state[37] <= 0;
						end else
						begin
							zx_Keyboard_state[37] <= 1;
						end 
															// C key, 0 pressed and 1 is released	
						if (r8Key1 == 8'h06 || r8Key2 == 8'h06 || r8Key3 == 8'h06) 
						begin
							zx_Keyboard_state[38] <= 0;
						end else
						begin
							zx_Keyboard_state[38] <= 1;
						end 	
															// V key, 0 pressed and 1 is released
						if (r8Key1 == 8'h19 || r8Key2 == 8'h19 || r8Key3 == 8'h19) //V
						begin
							zx_Keyboard_state[39] <= 0;
						end else
						begin
							zx_Keyboard_state[39] <= 1;
						end 
															//	Space, symb, MNB
															// Space key, 0 pressed and 1 is released
						if (r8Key1 == 8'h2c || r8Key2 == 8'h2c || r8Key3 == 8'h2c) 
						begin
							zx_Keyboard_state[30] <= 0;
						end else
						begin
							zx_Keyboard_state[30] <= 1;
						end 	
															// M key, 0 pressed and 1 is released
						if (r8Key1 == 8'h10 || r8Key2 == 8'h10 || r8Key3 == 8'h10) 
						begin
							zx_Keyboard_state[32] <= 0;
						end else
						begin
							zx_Keyboard_state[32] <= 1;
						end 	
															// N key, 0 pressed and 1 is released
						if (r8Key1 == 8'h11 || r8Key2 == 8'h11 || r8Key3 == 8'h11) 
						begin
							zx_Keyboard_state[33] <= 0;
						end else
						begin
							zx_Keyboard_state[33] <= 1;
						end 	
															// B key, 0 pressed and 1 is released
						if (r8Key1 == 8'h05 || r8Key2 == 8'h05 || r8Key3 == 8'h05) 
						begin
							zx_Keyboard_state[34] <= 0;
						end else
						begin
							zx_Keyboard_state[34] <= 1;
						end 	
															// ASDFG
															// A key, 0 pressed and 1 is released
						if (r8Key1 == 8'h04 || r8Key2 == 8'h04 || r8Key3 == 8'h04) 
						begin
							zx_Keyboard_state[25] <= 0;
						end else
						begin
							zx_Keyboard_state[25] <= 1;
						end 	
															// S key, 0 pressed and 1 is released
						if (r8Key1 == 8'h16 || r8Key2 == 8'h16 || r8Key3 == 8'h16) 
						begin
							zx_Keyboard_state[26] <= 0;
						end else
						begin
							zx_Keyboard_state[26] <= 1;
						end 	
															// D key, 0 pressed and 1 is released
						if (r8Key1 == 8'h07 || r8Key2 == 8'h07 || r8Key3 == 8'h07) 
						begin
							zx_Keyboard_state[27] <= 0;
						end else
						begin
							zx_Keyboard_state[27] <= 1;
						end 	
															// F key, 0 pressed and 1 is released
						if (r8Key1 == 8'h09 || r8Key2 == 8'h09 || r8Key3 == 8'h09) 
						begin
							zx_Keyboard_state[28] <= 0;
						end else
						begin
							zx_Keyboard_state[28] <= 1;
						end 	
															// G key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0A || r8Key2 == 8'h0A || r8Key3 == 8'h0A) 
						begin
							zx_Keyboard_state[29] <= 0;
						end else
						begin
							zx_Keyboard_state[29] <= 1;
						end 	

															// HJKL ENT
															// H key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0B || r8Key2 == 8'h0B || r8Key3 == 8'h0B) 
						begin
							zx_Keyboard_state[24] <= 0;
						end else
						begin
							zx_Keyboard_state[24] <= 1;
						end 	
															// J key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0D || r8Key2 == 8'h0D || r8Key3 == 8'h0D) 
						begin
							zx_Keyboard_state[23] <= 0;
						end else
						begin
							zx_Keyboard_state[23] <= 1;
						end 	
															// K key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0E || r8Key2 == 8'h0E || r8Key3 == 8'h0E) 
						begin
							zx_Keyboard_state[22] <= 0;
						end else
						begin
							zx_Keyboard_state[22] <= 1;
						end 	
															// L key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0F || r8Key2 == 8'h0F || r8Key3 == 8'h0F) 
						begin
							zx_Keyboard_state[21] <= 0;
						end else
						begin
							zx_Keyboard_state[21] <= 1;
						end 	
															// Enter key, 0 pressed and 1 is released
						if (r8Key1 == 8'h28 || r8Key2 == 8'h28 || r8Key3 == 8'h28) 
						begin
							zx_Keyboard_state[20] <= 0;
						end else
						begin
							zx_Keyboard_state[20] <= 1;
						end 	

															// QWERT
															// Q key, 0 pressed and 1 is released
						if (r8Key1 == 8'h14 || r8Key2 == 8'h14 || r8Key3 == 8'h14) 
						begin
							zx_Keyboard_state[15] <= 0;
						end else
						begin
							zx_Keyboard_state[15] <= 1;
						end 	
															// W key, 0 pressed and 1 is released
						if (r8Key1 == 8'h1a || r8Key2 == 8'h1a || r8Key3 == 8'h1a) 
						begin
							zx_Keyboard_state[16] <= 0;
						end else
						begin
							zx_Keyboard_state[16] <= 1;
						end 	
															// E key, 0 pressed and 1 is released
						if (r8Key1 == 8'h08 || r8Key2 == 8'h08 || r8Key3 == 8'h08) //E
						begin
							zx_Keyboard_state[17] <= 0;
						end else
						begin
							zx_Keyboard_state[17] <= 1;
						end 	
															// R key, 0 pressed and 1 is released
						if (r8Key1 == 8'h15 || r8Key2 == 8'h15 || r8Key3 == 8'h15) 
						begin
							zx_Keyboard_state[18] <= 0;
						end else
						begin
							zx_Keyboard_state[18] <= 1;
						end 	
															// T key, 0 pressed and 1 is released
						if (r8Key1 == 8'h17 || r8Key2 == 8'h17 || r8Key3 == 8'h17) 
						begin
							zx_Keyboard_state[19] <= 0;
						end else
						begin
							zx_Keyboard_state[19] <= 1;
						end 	

															// YUIOP
															// Y key, 0 pressed and 1 is released	
						if (r8Key1 == 8'h1c || r8Key2 == 8'h1c || r8Key3 == 8'h1c) 
						begin
							zx_Keyboard_state[14] <= 0;
						end else
						begin
							zx_Keyboard_state[14] <= 1;
						end 	
															// U key, 0 pressed and 1 is released
						if (r8Key1 == 8'h18 || r8Key2 == 8'h18 || r8Key3 == 8'h18) 
						begin
							zx_Keyboard_state[13] <= 0;
						end else
						begin
							zx_Keyboard_state[13] <= 1;
						end 	
															// I key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0c || r8Key2 == 8'h0c || r8Key3 == 8'h0c) 
						begin
							zx_Keyboard_state[12] <= 0;
						end else
						begin
							zx_Keyboard_state[12] <= 1;
						end 	
															// O key, 0 pressed and 1 is released
						if (r8Key1 == 8'h12 || r8Key2 == 8'h12 || r8Key3 == 8'h12) 
						begin
							zx_Keyboard_state[11] <= 0;
						end else
						begin
							zx_Keyboard_state[11] <= 1;
						end 	
															// P key, 0 pressed and 1 is released
						if (r8Key1 == 8'h13 || r8Key2 == 8'h13 || r8Key3 == 8'h13) 
						begin
							zx_Keyboard_state[10] <= 0;
						end else
						begin
							zx_Keyboard_state[10] <= 1;
						end 	
															// 12345
															// 5 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h22 || r8Key2 == 8'h22 || r8Key3 == 8'h22) 
						begin
							zx_Keyboard_state[9] <= 0;
						end else
						begin
							zx_Keyboard_state[9] <= 1;
						end 	
															// 4 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h21 || r8Key2 == 8'h21 || r8Key3 == 8'h21) 
						begin
							zx_Keyboard_state[8] <= 0;
						end else
						begin
							zx_Keyboard_state[8] <= 1;
						end 	
															// 3 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h20 || r8Key2 == 8'h20 || r8Key3 == 8'h20) 
						begin
							zx_Keyboard_state[7] <= 0;
						end else
						begin
							zx_Keyboard_state[7] <= 1;
						end 	
															// 2 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h1f || r8Key2 == 8'h1f || r8Key3 == 8'h1f) 
						begin
							zx_Keyboard_state[6] <= 0;
						end else
						begin
							zx_Keyboard_state[6] <= 1;
						end 	
															// 1 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h1e || r8Key2 == 8'h1e || r8Key3 == 8'h1e) 
						begin
							zx_Keyboard_state[5] <= 0;
						end else
						begin
							zx_Keyboard_state[5] <= 1;
						end 	
															// 67890
															// 6 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h23 || r8Key2 == 8'h23 || r8Key3 == 8'h23) 
						begin
							zx_Keyboard_state[4] <= 0;
						end else
						begin
							zx_Keyboard_state[4] <= 1;
						end 	
															// 7 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h24 || r8Key2 == 8'h24 || r8Key3 == 8'h24) 
						begin
							zx_Keyboard_state[3] <= 0;
						end else
						begin
							zx_Keyboard_state[3] <= 1;
						end 	
															// 8 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h25 || r8Key2 == 8'h25 || r8Key3 == 8'h25) 
						begin
							zx_Keyboard_state[2] <= 0;
						end else
						begin
							zx_Keyboard_state[2] <= 1;
						end 	
															// 9 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h26 || r8Key2 == 8'h26 || r8Key3 == 8'h26) 
						begin
							zx_Keyboard_state[1] <= 0;
						end else
						begin
							zx_Keyboard_state[1] <= 1;
						end 	
															// 0 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h27 || r8Key2 == 8'h27 || r8Key3 == 8'h27) 
						begin
							zx_Keyboard_state[0] <= 0;
						end else
						begin
							zx_Keyboard_state[0] <= 1;
						end 	
					end
															// Increment bytes counter
					rx_byte_cntr = rx_byte_cntr + 1;
				end
				else begin								// If we are outside the packet then reset bytes counter
					rx_byte_cntr = 0;
				end
				r8Uart115RxDelay <= 216;         // One more delay will be required              
				r8Uart115RxState <= 255;         // So jump to delay state                       
				r8Uart115RxNextState <= 0;       // And set next counter to 0                    
			end 
			255:                                // Delay state                                  
			begin
				if (r8Uart115RxDelay != 0)       // If delay counter neq 0                       
				begin   	
															// Then decrement it
					r8Uart115RxDelay <= r8Uart115RxDelay - 1;                            
					if (bRx115ShiftReq)           // If shift was requested                       
					begin
						bRx115ShiftReq <= 0;       // Set shift req flag to 0                      
                                             // Actual shift                                 
						r8Uart115RxByte <= r8Uart115RxByte >> 1;   
					end
				end else                         // Otherwise                                    
				begin
															// Jump to the next required state              
					r8Uart115RxState <= r8Uart115RxNextState;
				end
			end
			default:                            // Preacution                                   
				begin
					r8Uart115RxState <= 0;        // State to 0                                   
				end
		endcase
	end
end


always@(negedge rCpuClkMain or negedge rst_n)// Interrupts on ZX Spectrum must happen everey 1/50S and last for 32 CPU clocks. Ideally they 
															// should be aligned with Vsync (which is 700000 of 10xCPU clocks) but in our case due to the fact that 
															//	Vsync = 60Hz (VGA timing) they are not aligned to actual Vsync but aligned to a virtual one. 
															// This feature itself is used for keyboard polling on ZX Spectrum
begin
	if (rst_n == 1'b0)
	begin
		a_nINT <= 1;									// Init interrupt flag
		sp50Hzcntr <= 0;								// Reset the counter
 		a_nWAIT <= 1;									// CPU signals, no wait			
 		a_nNMI <= 1;									//	CPU signals, no unmaskable interrupt		
 		a_nBUSRQ <= 1;									// CPU signals, no BUSREQ
		PreviousVsync <= 0;
	end else
	begin 
		if (outVsync == 0 && PreviousVsync == 1)
		begin
			sp50Hzcntr <= 0;							// Clear counter
			PreviousVsync <= 0;						// Clear previous vsync flag
		end else
		begin
			if (outVsync == 1)						// Wait till frame is active 
			begin		
				PreviousVsync <= 1;					// Set Vsync flag so we will know in the lines above if Vsync happen				
			end
															// If current pixel number is greater than 850 (VIDEO_PIX_INT) 
			if (outLineCntr * `HTOTAL + outPixCntr > `VIDEO_PIX_INT)
			begin
				case (sp50Hzcntr)						// Interrup shall take 31 CPU clocks
					0:  begin
						a_nINT <= 0;					// Interrupt flag set			
															// Increment counter
						sp50Hzcntr <= sp50Hzcntr + 1;		
					end
					31: begin							// Time to clear the interrupt
						a_nINT <= 1;					// Clear interrupt flag
															// Increment counter
						sp50Hzcntr <= sp50Hzcntr + 1;		
					end
					default: begin
															// Increment counter
						sp50Hzcntr <= sp50Hzcntr + 1;		
					end
				endcase
			end
		end
	end
end


// End of ZX Spectrum 3.5MHz clock generator

always @(*)												// 7 segment led decoder. When any of the signal to be displayed changes
begin
	if (rst_n == 1'b0)								// If we are in reset state
	begin
		tstDATA_LED <= 0;								// Reset LEDs to 0
	end else
	begin
		case({rLED_Special, rLED_Bcd})         // Otherwise depending on what has been requsted
			5'b00000: tstDATA_LED<= 8'b11000000;// "0"     
			5'b00001: tstDATA_LED<= 8'b11111001;// "1" 
			5'b00010: tstDATA_LED<= 8'b10100100;// "2" 
			5'b00011: tstDATA_LED<= 8'b10110000;// "3" 
			5'b00100: tstDATA_LED<= 8'b10011001;// "4" 
			5'b00101: tstDATA_LED<= 8'b10010010;// "5" 
			5'b00110: tstDATA_LED<= 8'b10000010;// "6" 
			5'b00111: tstDATA_LED<= 8'b11011000;// "7" 
			5'b01000: tstDATA_LED<= 8'b10000000;// "8"    
			5'b01001: tstDATA_LED<= 8'b10010000;// "9" 
			5'b01010: tstDATA_LED<= 8'b10001000;// "A" 
			5'b01011: tstDATA_LED<= 8'b10000011;// "B" 
			5'b01100: tstDATA_LED<= 8'b11000110;// "C" 
			5'b01101: tstDATA_LED<= 8'b10100001;// "D" 
			5'b01110: tstDATA_LED<= 8'b10000110;// "E" 
			5'b01111: tstDATA_LED<= 8'b10001110;// "F" 
			5'b10000: tstDATA_LED<= 8'b10111111;// "-" Special thanks to MK61 and MK52
			5'b10001: tstDATA_LED<= 8'b10000000;// "8" Special thanks to MK61 and MK52
			5'b10010: tstDATA_LED<= 8'b11000110;// "C" Special thanks to MK61 and MK52
			5'b10011: tstDATA_LED<= 8'b10000110;// "E  Special thanks to MK61 and MK52
			5'b10100: tstDATA_LED<= 8'b11111111;// " " Special thanks to MK61 and MK52
			5'b10101: tstDATA_LED<= 8'b10101111;// "r" Special thanks to MK61 and MK52
			5'b10110: tstDATA_LED<= 8'b10100011;// "o" Special thanks to MK61 and MK52
			default: tstDATA_LED <= 8'b11000000;// "0"
		endcase
	end
end

always @(negedge clk or negedge rst_n)			//  Reducing a speed for led to 25MHz
begin
	if (rst_n == 1'b0)								// In case of reset
	begin
		clk25 <= 0;										// Assign clk 25 to 0
	end else
	begin
		clk25 =~clk25;									// Invert clk25
	end
end
	 
always@(posedge clk25 or negedge rst_n)   	// 6 LED, 7 segments
begin
	if (rst_n == 1'b0)
	begin
		tstDATA_LED_SEL <= 6'b111111;				// LED select at reset eq 'all off'
		rLED_Mcntr <= 0;								// Timer at reset eq 0
		rLED_Sel <= 0;									// Binary LED selector
	end else
	begin
		if (rLED_Mcntr != 0)							// If timer not equal 0
		begin
			rLED_Mcntr <= rLED_Mcntr + 1;			// The simply increment it
		end else
		begin
			rLED_Mcntr <= 1;							// Otherrwise if it 0, then set it to 1
			case (rLED_Sel)							// Then proceed to led selector
				0: begin									// LSB LED
					tstDATA_LED_SEL <= 6'b011111; // Select LED #0 (lsb LED)
					rLED_Bcd <= rLED_Data[3:0];	// Load data for it
					rLED_Sel <= 1;						// Then we are going to update next LED
				end
				1: begin									// Next LED
					tstDATA_LED_SEL <= 6'b101111; // Select LED #1
					rLED_Bcd <= rLED_Data[7:4];   // Load data for it
					rLED_Sel <= 2;						// Then we are going to update next LED
				end
				2: begin									// Next LED
					tstDATA_LED_SEL <= 6'b110111; // Select LED #2
					rLED_Bcd <= rLED_Data[11:8];  // Load data for it
					rLED_Sel <= 3;						// Then we are going to update next LED	
				end
				3: begin                         // Next LED
					tstDATA_LED_SEL <= 6'b111011; // Select LED #3
					rLED_Bcd <= rLED_Data[15:12]; // Load data for it
					rLED_Sel <= 4;						// Then we are going to update next LED	
				end
				4: begin									// Next LED
					tstDATA_LED_SEL <= 6'b111101; // Select LED #4
					rLED_Bcd <= rLED_Data[19:16]; // Load data for it
					rLED_Sel <= 5;						// Then we are going to update next LED	
				end
				5: begin									// Next LED
					tstDATA_LED_SEL <= 6'b111110; // Select LED #5
					rLED_Bcd <= rLED_Data[23:20]; // Load data for it
					rLED_Sel <= 0;						// Then we are going to update next LED	
				end
				default: begin							// Default state to avoid any potential issue
					tstDATA_LED_SEL <= 6'b111111; // All LEDs are off
					rLED_Bcd <= rLED_Data[7:0];	// Load some data	
					rLED_Sel <= 0;						// Set the selector to 0 to go thru the first state
				end
			endcase	
		end
	end
end

always@(posedge m56clk or negedge rst_n)     // Memory controller
//always@(negedge m28clk or negedge rst_n)     // Memory controller
begin
	if (rst_n == 1'b0)                        // In case of reset
	begin
		ex_nWR_EN <= 1;                        // Set write enable to read from RAM
		ext_n_CE <= 0;					 				// Set CE to 0
		ext_n_OE <= 0;									// Set OE to 0
		rBorderColor <= 2;							// At reset set border to be 2
		sound <= 1;										// Set sound enable bit
		bPrevSPI <= 0; 								// Bit indicating previous state of SPI transfer
		bZ80Reset <= 0;								// CPU Reset
		aROMCopyAddr <= 0;							// Copy address for the ROM for SNOR copying
		b2RomCopySM <= 0;								// State machine for SNOR copy	
		v32CheckSum <= 0;								// Checksum for SNOR
		rLED_Data <= 24'h00000;						// Debug
		rLED_Special <= 0;							// Debug
		r8_128KSwitch <= 0;							// Memory switch byte is equal 0 at strat
	end else
	begin
	
		if (bZ80Reset == 1)
		begin
			if (a_nIORQ == 0)							// If we are in the port access mode			
			begin
				
				if (a_nWR == 0)						// If we are writing then on ZX spectrum the only port to write is 0xFE, there are no other ports
				begin
															// If we are writing to a port 0xFE (see ZX spectrum documentation) 
					if (wZXGlobalAddress[7:0] == 8'hfe)															
					begin
															// Set the border color, bits 2:0
						rBorderColor[2:0] <= a_D[2:0];
						sound <= a_D[4];				// Set the sound bit, bit 4
					end else
					begin									// As normal on Sinclair hardware, the port address is in fact only partially decoded 
															// and the hardware will respond to any port address with bits 1 and 15 reset
						if (wZXGlobalAddress[15] == 0 && wZXGlobalAddress[1] == 0 && r8_128KSwitch[5] == 0)
						begin
							r8_128KSwitch <= a_D;	// Update memory control from CPU bus
							rLED_Data <= a_D;
						end else
						begin
							if (wZXGlobalAddress[15] == 1 && wZXGlobalAddress[1] == 0)
							begin
								AY8910_d_in <= a_D;	// AY8910 write
								led[3] = ~led[3];
							end
						end
					end	
				end else
				begin
					if (a_nRD == 0)					// If we are reading port 
					begin
						
															// If it is AY8910
						if (wZXGlobalAddress[15] == 1 && wZXGlobalAddress[1] == 0)
						begin
															// Then provide AY8910 data
							wZXGlobalDataO[7:0] <= AY8910_d_out;
							led[2] = ~led[2];
						end else
						begin
						
							case (wZXGlobalAddress)	// We need to see which one we are reading
								16'hfefe: begin		// Port fefe - keyobard inputs, caps thru Z
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[39:35];
								end						// Port 7ffe
								16'h7ffe: begin		// Keys BNM, sym shift and space
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[34:30];
								end
								16'hfdfe: begin		// Keys ASDFG
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[29:25];
								end
								16'hbffe: begin		// Keys HJKL Enter
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[24:20];
								end
								16'hfbfe: begin		// Keys QWERT
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[19:15];
								end
								16'hdffe: begin		// Keys YUIOP
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[14:10];
								end
								16'hf7fe: begin		// Keys 12345
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[9:5];
								end
								16'heffe: begin		// Keys 67890
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[4:0];
								end
								default: begin			// "ANY" key simulation in case if program polls port 254 as indication of such event
															// If no key pressed
									if (zx_Keyboard_state == 40'hffffffffff)
									begin
															// ...then load all ones
										wZXGlobalDataO[4:0] <= 5'b11111;
									end else 
									begin
															// Otherwise simulate "any" key
										wZXGlobalDataO[4:0] <= 5'b11110;
									end
								end
							endcase
															// Port FE, see ZX Spectrum manual
							if (wZXGlobalAddress[7:0] == 8'hFE)
							begin
															// Bit 5 is not used
								wZXGlobalDataO[5] <= 1;
															// Bit 6 is input from a tape
								wZXGlobalDataO[6]<= ~tapein;
															// Bit 7 is not used
								wZXGlobalDataO[7] <= 1;		
							end else
															// Kempston joystick	
															// bit D0:     right            PIN_J1_30    PIN_FPGA_L9    PIN_JSTK_04
															// bit D1:     left             PIN_J1_32    PIN_FPGA_P8    PIN_JSTK_03
															// bit D2:     down             PIN_J1_34    PIN_FPGA_P2    PIN_JSTK_02
															// bit D3:     up               PIN_J1_36    PIN_FPGA_N2    PIN_JSTK_01
															// bit D4:     fire             PIN_J1_28    PIN_FPGA_M9    PIN_JSTK_06
															// bits D5-D7: always zero	
								if (wZXGlobalAddress[7:0] == 8'h1F)
								begin
															// MSBs are eq 0
									wZXGlobalDataO[7:5] <= 3'b000;
															// Bits are inverted
									wZXGlobalDataO[4:0] <= ~KempstonJoystick[4:0];
								end else
								begin						// Special feature to emulate ULA memory leakage
									if (bDrawingBorder == 0)
									begin					// In the active area non used ports shall return current byte being displayed
										if (r8_128KSwitch[3] == 0)
										begin					// Active page is 5
											wZXGlobalDataO[7:0] <= aVRAM_q_a;
										end else
										begin					// Active page is 7
											wZXGlobalDataO[7:0] <= bVRAM_q_a;
										end
									end else
									begin
															// While it should return FF if drawing the border
										wZXGlobalDataO[7:0] <= 8'hFF;
									end
								end
						end 
					end
				end
			end else
			begin
															// If address is less than 16384 the this is ROM access
				if (wZXGlobalAddress < `ZX_RAM_START)
				begin 
					ex_nWR_EN <= 1;					// For ROM write enable is always inactive
					wZXGlobalDataO<= w_extDATA;	// Load external data to a variable
															// Rom address calculation
					if (r8_128KSwitch[4] == 1)		// If bit 4 is set then it is standard 48K ROM, started at address 16384
					begin
						extADDR <= wZXGlobalAddress + `ROM2ND_OFFSET;
					end else
					begin									// Otherwise it is 128K ROM
						extADDR <= wZXGlobalAddress;
					end
				end else
				begin										// If address is less than ZX_RAM_START the it is lower RAM access
															// screeen$
					if (wZXGlobalAddress < `ZX_RAM_START + 8192)
					begin
															// Set the address as requested address less ZX_RAM_START
															// As RAM block starts internally from 0	
						aVRAM_addr_b <= wZXGlobalAddress - `ZX_RAM_START;
						aVRAM_wren_b <= ~a_nWR;		// Set wr_enable for embedded RAM
															// Data from embdded RAM to be set as data output of this module
						wZXGlobalDataO <= aVRAM_q_b;	
						aVRAM_data_b <= a_D;  		// Data to embedded RAM is set from the data to this module
					end
				
					if (wZXGlobalAddress >= `ZX_RAM_START + 8192 && wZXGlobalAddress < 16'h8000)
					begin
															// Main RAM, page 5 access, address is 32768 or above
						ex_nWR_EN <= a_nWR;			// Set read/write flag as requested from outside
															// Set the data out from this module to a data bus from external RAM module
						wZXGlobalDataO <= w_extDATA;	
															// Set address bus less 32768 but on top of ROM
															// ZX Ram address is shifted by 32768 as we will keep lower 32K for two paged ROM
						//r8_128KSwitch
						extADDR <= wZXGlobalAddress + 8192 + 5*16384;
															// Set the data to external RAM module to the one which is being passed
						rExternal_mem_data_bus <= a_D;
					end
				
					if (wZXGlobalAddress >= 16'h8000 && wZXGlobalAddress < 16'hc000)
					begin
															// Main RAM, page 2 access 
						ex_nWR_EN <= a_nWR;			// Set read/write flag as requested from outside
															// Set the data out from this module to a data bus from external RAM module
						wZXGlobalDataO <= w_extDATA;	
															// Set address bus less 32768 but on top of ROM
															// ZX Ram address is shifted by 32768 as we will keep lower 32K for two paged ROM
						//r8_128KSwitch
						extADDR <= wZXGlobalAddress + 2*16384;
															// Set the data to external RAM module to the one which is being passed
						rExternal_mem_data_bus <= a_D;
					
					end
					
					if (wZXGlobalAddress >= 16'hc000)
					begin
						case (r8_128KSwitch[2:0])
							0, 1, 2,	3,	4,	6: begin
															// Main RAM, page XX access 
								ex_nWR_EN <= a_nWR;	// Set read/write flag as requested from outside
															// Set the data out from this module to a data bus from external RAM module
								wZXGlobalDataO <= w_extDATA;	
															// Set address bus less 32768 but on top of ROM
															// ZX Ram address is shifted by 32768 as we will keep lower 32K for two paged ROM
								extADDR <= wZXGlobalAddress - 16384 + r8_128KSwitch[2:0]*16384;
															// Set the data to external RAM module to the one which is being passed
								rExternal_mem_data_bus <= a_D;
							end
							5: begin
								if (wZXGlobalAddress < 16'he000)
								begin
															// Set the address as requested address less ZX_RAM_START
															// As RAM block starts internally from 0	
									aVRAM_addr_b <= wZXGlobalAddress -  16'hC000;
															// Set wr_enable for embedded RAM
									aVRAM_wren_b <= ~a_nWR;		
															// Data from embdded RAM to be set as data output of this module
									wZXGlobalDataO <= aVRAM_q_b;	
															// Data to embedded RAM is set from the data to this module
									aVRAM_data_b <= a_D;  		
								end else
								begin
															// Main RAM, page XX access 
															// Set read/write flag as requested from outside
									ex_nWR_EN <= a_nWR;	
															// Set the data out from this module to a data bus from external RAM module
									wZXGlobalDataO <= w_extDATA;	
															// Set address bus less 32768 but on top of ROM
															// ZX Ram address is shifted by 32768 as we will keep lower 32K for two paged ROM
									extADDR <= wZXGlobalAddress - 16384 + r8_128KSwitch[2:0]*16384;
															// Set the data to external RAM module to the one which is being passed
									rExternal_mem_data_bus <= a_D;
								end
							end
							7: begin
								if (wZXGlobalAddress < 16'he000)
								begin
															// Set the address as requested address less ZX_RAM_START
															// As RAM block starts internally from 0	
									bVRAM_addr_b <= wZXGlobalAddress - 16'hC000;
															// Set wr_enable for embedded RAM
									bVRAM_wren_b <= ~a_nWR;		
															// Data from embdded RAM to be set as data output of this module
									wZXGlobalDataO <= bVRAM_q_b;	
															// Data to embedded RAM is set from the data to this module
									bVRAM_data_b <= a_D;  		
								end else
								begin
															// Main RAM, page XX access 
															// Set read/write flag as requested from outside
									ex_nWR_EN <= a_nWR;	
															// Set the data out from this module to a data bus from external RAM module
									wZXGlobalDataO <= w_extDATA;	
															// Set address bus less 32768 but on top of ROM
															// ZX Ram address is shifted by 32768 as we will keep lower 32K for two paged ROM
									extADDR <= wZXGlobalAddress - 16384 + r8_128KSwitch[2:0]*16384;
															// Set the data to external RAM module to the one which is being passed
									rExternal_mem_data_bus <= a_D;
								end
							end
						endcase
					
					end
					
				end	
			end 
		end else
		begin
			if (bPrevSPI != bSPIind)				// To sync with SPI loader running at different clock domain we use flipping flag. 
															// When new byte is available from SPI loader then this flag is flipped
			begin
				case (b2RomCopySM)					// State machine for copying the data
					0: 
					begin
						extADDR <= aROMCopyAddr;	// Load address to the address bus
						ex_nWR_EN <= 1;				// Set RD/WR flag
						ext_n_OE <= 1;					// Set OE flag	
															// Set the data
						rExternal_mem_data_bus <= V8_DIr;
						b2RomCopySM = 1;				// Increment state machine
					end
					1:
					begin
						ex_nWR_EN <= 0;				// Set RD/WR flag	
						ext_n_OE <= 1;					// Set OE = 1
						b2RomCopySM = 2;				// Increment the state
					end
					2:
					begin
						b2RomCopySM = 3;				// Just wait in this state
					end
					3: 
					begin
						ex_nWR_EN <= 1;				// Set RD/WR flag
						ext_n_OE <= 0;					// Set OE flag	
						b2RomCopySM = 0;				// Reset the state machine
						bPrevSPI <= bSPIind;			// Copy current state of SPI data ready flag, so next time we will get in only when it changes
															// Add current byte to a checksum
						v32CheckSum = v32CheckSum + V8_DIr;
						rLED_Data = v32CheckSum + V8_DIr;
															// Increment an address
						aROMCopyAddr = aROMCopyAddr + 1;
					end
				endcase
			end else
			begin
				if (bSPI_done == 1)					// If SPI loading is done
				begin
					ext_n_OE <= 0;						// set OE flag
					bZ80Reset <= 1;					// Release Z80 from reset
					rBorderColor <= 1;
					r8_128KSwitch <= 0;				// Memory switch reg should be 0 at all times during reset
					rLED_Data <= v32CheckSum;		// Set the checksum to be displayed on LED
				end 
			end
		end
	end 
end

always@(posedge outVsync or negedge rst_n)	// Flash counter, runs on Vsync
begin
	if (rst_n == 1'b0)								// If reset is active
	begin
		rZXFlashAttr <= 0;							// Set flash attrbute to 0
		rZXFlashCntr <= 0;							// Set flash counter to 0	
	end else
	begin
		if (rZXFlashCntr > `FLASHSPEED)			// If we reached 19 (flash speed)
		begin
			rZXFlashCntr <= 0;						// Clear flash counter
			rZXFlashAttr <= ~rZXFlashAttr;		// Invert flash attribute	
		end else
		begin
			rZXFlashCntr <= rZXFlashCntr + 1;   // If we did not reach end, then increment the counter
		end
	end
end
															// Main process for video
always@(posedge clk or negedge bZ80Reset)
begin
															// If reset is active
	if (bZ80Reset == 1'b0)								
	begin
		outVideoPCLK <= 0; 							// Set PCLK out to 0
		outPixCntr <= 0;								// Set Pixels in line counter to 0
		outVsync <= 0;									// Set output vsync to 0
		outHsync <= 0;									// Set output hsync to 0
		outLineCntr <= 0;								// Set lines in frame counter to 0
		out_B <= 0;										// Set B default to 0
		out_G <= 0;										// Set G default to 0
		out_R <= 0;										// Set R default to 0
		aVRAM_data_a <= 0;							// Set video data in from videoadater to 0
		aVRAM_wren_a <= 0;							// Set write enable from videoadater to 0
		aVRAM_addr_a <= 0;							// Set video adatper address to 0	
		bVRAM_data_a <= 0;							// Set video data in from videoadater to 0
		bVRAM_wren_a <= 0;							// Set write enable from videoadater to 0
		bVRAM_addr_a <= 0;							// Set video adatper address to 0	
		pixeladdr <= 0;								// Set plain pixel address (not in ZX way) to 0
		ColorLineAddr <= 0;							// Set color attribute counter (1 line only), range 0..31, used in calculating VRAM memory address
		ColorSync <= 0;								// Set flag to indicate even/odd pixel cylcle to give enough timeout to fetch data from color VRAM
		ColorDataMemCpy <= 0;						// Set color line (256 bits) to 0
		CurColorAttribute <= 0;						// Set current color attribute to 0
		CurColorLineAddr <= 0;						// Set address for attrubute cache to 0
		bDrawingBorder <= 0;
	end else												// if NOT reset
	begin												
		if (outVideoPCLK == 1)						// We need to devide 50M osc by 2 to get to 25M and assign active edge, so it will be H->L
		begin
			outVideoPCLK <= 0;						// Clear output PCLK
			case (outLineCntr)						// Check if it is time for Vsync
				`VSYNCSTART: begin					// If we are at the beggining of the frame then set Vsync to 0 as polarity is NEG
					outVsync <= 0;
				end
				`VSYNCEND: begin						// If we are at the end of the frame sync pulse then set Vsync to 1
					outVsync <= 1;
				end
			endcase
															// If we are in the active frame
			if (outPixCntr >= `HACTSTART && outPixCntr < `HACTEND && outLineCntr >= `VACTSTART && outLineCntr < `VACTEND)
			begin
															// ZX Spectrum screen is 256x192px, double that and it gives us 512x324 px
															// Hbackporch = (640-512)/2 = 64; Vbackporch = (480-324)/2 = 78
															// If we are at the edge the draw while rectangle  35+78  till 515 - 78  ie 384
				if (outPixCntr >= `HACTSTART + `HBACKPORCH && outPixCntr < `HACTEND - `HBACKPORCH && outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER)
				begin
					case (outPixCntr[3:0])			// If lsb pixel[2:0]*2 clocks == 0 then it means that we are at the beggining of new color space (8 pixels) on the screen
						0: begin
															// Fetching color attribute
							CurColorAttribute = ColorDataMemCpy[255:248];
						end
						1: begin
															// Fetching color attribute
							ColorDataMemCpy = ColorDataMemCpy << 8;
						end
					endcase
															// INK color
					if ((r8_128KSwitch[3] == 0 &&(aVRAM_q_a [7 - outPixCntr[3:1]] == !(rZXFlashAttr & CurColorAttribute[7]))) || (r8_128KSwitch[3] == 1 &&(bVRAM_q_a [7 - outPixCntr[3:1]] == !(rZXFlashAttr & CurColorAttribute[7]))))
					begin
															// Separate R, G and B, so this one is BLUE
						if ((CurColorAttribute & 1) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin
								out_B <= 5'b11111;	// Load bright blue
							end else
							begin
								out_B <= 5'b10111;   // Load darker blue
							end
						end else
						begin
							out_B <= 0;			      // If there is no blue, then load blue = 0
						end
															// Separate R, G and B, so this one is RED
						if ((CurColorAttribute & 2) > 0)
						begin								
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin
								out_R <= 5'b11111;	// Load bright red
							end else
							begin
								out_R <= 5'b10111;	// Load darker red
							end
						end else
						begin
							out_R <= 0;       		// If there is no red, then load red = 0
						end
															// Separate R, G and B, so this one is GREEN
						if ((CurColorAttribute & 4) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin 
								out_G <= 6'b111111;	// Load bright green		
							end else
							begin
								out_G <= 6'b110010;	// Load darker green		
							end
						end else
						begin
							out_G <= 0;					// If there is no green, then load green = 0;
						end
						
					end else
					begin
															// PAPER color
															// Separate R, G and B, so this one is BLUE
						if ((CurColorAttribute & 8) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin
								out_B <= 5'b11111;   // Load bright blue
							end else
							begin
								out_B <= 5'b10111;   // Load darker blue
							end
						end else
						begin
							out_B <= 0;			      // If there is no blue, then 0
						end
															// Separate R, G and B, so this one is RED
						if ((CurColorAttribute & 16) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin
								out_R <= 5'b11111;	// Load bright red
							end else
							begin
								out_R <= 5'b10111;	// Load darker red		
							end
						end else
						begin
							out_R <= 0;					// If there is no red, then load 0
						end
															// Separate R, G and B, so this one is GREEN
						if ((CurColorAttribute & 32) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin 
								out_G <= 6'b111111;  // Set bright green
							end else
							begin
								out_G <= 6'b110010;	// Set darker green		
							end
						end else
						begin
							out_G <= 0;			      // If there is no gree, then set it to 0
						end
					end
					bDrawingBorder <= 0;	
				end else
				begin
															// Border color
					bDrawingBorder <= 1;	
					if (rBorderColor[0] == 1)		// Bit 0 is blue
					begin
						out_B <= 5'b10111;			// If it is set, then program out B to non-full brightness			
					end else
					begin
						out_B <= 5'b00000;			// If it is clear then program out B as no color			
					end
					
					if (rBorderColor[1] == 1) 		// Bit 1 is red
					begin
						out_R <= 5'b10111;			// If it is set, then program out R to non-full brightness			
					end else
					begin
						out_R <= 0;						// If it is clear then program out R as no color			
					end
					
					if (rBorderColor[2] == 1)		// Bit 2 is green
					begin
						out_G <= 6'b110010;			// If it is set, then program out G to non-full brightness			
					end else
					begin
						out_G <= 0;						// If it is clear then program out G as no color			
					end
					
				end
			end else
			begin
				out_B <= 0;								// If we are in the balnking area then assign B = 0
				out_G <= 0;								// If we are in the balnking area then assign G = 0
				out_R <= 0;								// If we are in the balnking area then assign R = 0
															// If we are within the active area vertically but in the backporch horizontally
															// then we need to pre-fetch color data from FPGA vram
				if (outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER && outPixCntr < `HACTSTART + `HBACKPORCH)
				begin
					if (ColorSync == 1)				// We need to do every other cycle to sync with FPGA RAM
					begin

						if (CurColorLineAddr < 32)
						begin
							if (r8_128KSwitch[3] == 0)
							begin
															// Load data from page 5
								ColorDataMemCpy = {ColorDataMemCpy[247:0], aVRAM_q_a}; 
							end else
							begin
															// Load data from page 7
								ColorDataMemCpy = {ColorDataMemCpy[247:0], bVRAM_q_a}; 
							end
															// And increment the pointer
							CurColorLineAddr = CurColorLineAddr + 1;
						end
					end 
				end 
			end											
															// Pre-fetching color data from FPGA RAM in vertical active area and a H backporch
			if (outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER && outPixCntr < `HACTSTART + `HBACKPORCH - 1)
			begin
				if (ColorSync == 0)					// If flag to sync (2x down clock speed) is eq 0
				begin
															// Calculate current address for color attribute
															
					if (r8_128KSwitch[3] == 0)    // r8_128KSwitch[3] tells us which video memory is used, 0 or 1, they are located in page 5 and 7
					begin
															// Set address for page 5
						aVRAM_addr_a = 6144 + ColorLineAddr + (((outLineCntr - `VACTSTART - `VBACKPORCH) >> 4)*32);
					end else
					begin
															// Set address for page 7
						bVRAM_addr_a = 6144 + ColorLineAddr + (((outLineCntr - `VACTSTART - `VBACKPORCH) >> 4)*32);
					end

					if (ColorLineAddr < 32)       // If there is next attrbute then
					begin
															// Increment current address
						ColorLineAddr = ColorLineAddr + 1;
					end
				end
			end else
			begin
				ColorLineAddr = 0;					// If we are out of color fetching area
				CurColorLineAddr = 0;				// Then set both counters, FPGA RAM address and 256 bit one to 0
			end
															//If NEXT pixel is in active area then setup address to read it
			if (outPixCntr >= `HACTSTART + `HBACKPORCH - 1 && outPixCntr < `HACTEND - `HBACKPORCH - 1 && outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER)
			begin
				pixeladdr = (((outPixCntr + 1 - `HACTSTART - `HBACKPORCH)>>1) >> 3) + (((outLineCntr - `VACTSTART - `VBACKPORCH) >> 1) * 32); 
															// Blocking assignments literally block the execution of the next
															// statement until the current statement is executed. (blocking '=', non blocking '<=')
															// â€“ Consequently, blocking assignments result in ordered
															// statement execution.
															// For example:
															// assume a = b = 0 initially;
															// a = 1; //executed first
															// b = a; //executed second
															// then a = 1, b = 1 after ordered execution. So:
															// ZX video memory is organized as following:
															// Y7  Y6  Y2  Y1  Y0 Y5  Y4 Y3 X4 X3 X2 X1 X0 ..or..
															// Y12 Y11 Y7  Y6  Y5 Y10 Y9 Y8 X4 X3 X2 X1 X0
				if (r8_128KSwitch[3] == 0)			// r8_128KSwitch[3] tells us which video memory is used, 0 or 1, they are located in page 5 and 7
				begin
															// Set address to page 5
					aVRAM_addr_a = pixeladdr[4:0] | pixeladdr[10:8] << 5 | pixeladdr[7:5] << 8 | pixeladdr [12:11] << 11;
				end else
				begin
															// Set address to page 7
					bVRAM_addr_a = pixeladdr[4:0] | pixeladdr[10:8] << 5 | pixeladdr[7:5] << 8 | pixeladdr [12:11] << 11;
				end 
			end

			case (outPixCntr)							// Check if it is time for Hsync
				`HSYNCSTART: begin					// If we are at the beginning of the line then set Hsync to 0 as polarity is NEG
					outHsync <= 0;
				end
				`HSYNCEND: begin						// If we are at the end of horizontal sync pulse then set Hsync to 1
					outHsync <= 1;
				end
			endcase
			
		end else											// If we are at the L->H PCLK satate
		begin
			ColorSync = ~ColorSync;
			outVideoPCLK = 1;							// Assign 1 to PCLK
			if (outPixCntr >= `HTOTAL - 1)		// If we are done with the line	
			begin
				outPixCntr <= 0;						// Set H pixel count to 0
				if (outLineCntr >= `VTOTAL - 1)	// If we are done with the frame
				begin
					outLineCntr <= 0;					// Set V counter to 0
				end else
				begin										// If not done then increment it
					outLineCntr <= outLineCntr + 1;
				end
			end else
			begin
				outPixCntr <= outPixCntr + 1;		// If we are not done with the line then increment it
			end
		end
	end
end

endmodule

